/*---------------------------------------------------------------------------

   wpng - PNG-writing program                                          wpng.c

  ---------------------------------------------------------------------------*/

#define PROGNAME  "wpng"
#define VERSION   "1.03 of 19 March 2000"
#define APPNAME   "Simple PGM/PPM/PAM to PNG Converter"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>     /* for jmpbuf declaration in writepng.h */
#include <time.h>


#include "writepng.h"   /* typedefs, common macros, writepng prototypes */


/* local prototypes */
static void wpng_cleanup(void);

static mainprog_info wpng_info;   /* lone global */

int wpng(char *data, int width, int height, int type, int maxval, FILE *outf)
{
    char *p;
    ulg rowbytes;
    int rc;
    int error = 0;
    double LUT_exponent;                /* just the lookup table */
    double CRT_exponent = 2.2;          /* just the monitor */
    double default_display_exponent;    /* whole display system */
    double default_gamma = 0.0;

    wpng_info.outfile = NULL;
    wpng_info.image_data = NULL;
    wpng_info.row_pointers = NULL;
    wpng_info.interlaced = TRUE;
    wpng_info.have_bg = FALSE;
    wpng_info.have_time = FALSE;
    wpng_info.have_text = 0;
    wpng_info.gamma = 0.0;


    /* First get the default value for our display-system exponent, i.e.,
     * the product of the CRT exponent and the exponent corresponding to
     * the frame-buffer's lookup table (LUT), if any.  If the PNM image
     * looks correct on the user's display system, its file gamma is the
     * inverse of this value.  (Note that this is not an exhaustive list
     * of LUT values--e.g., OpenStep has a lot of weird ones--but it should
     * cover 99% of the current possibilities.  This section must ensure
     * that default_display_exponent is positive.) */

#if defined(NeXT)
    /* third-party utilities can modify the default LUT exponent */
    LUT_exponent = 1.0 / 2.2;
#elif defined(sgi)
    LUT_exponent = 1.0 / 1.7;
    /* there doesn't seem to be any documented function to
     * get the "gamma" value, so we do it the hard way */
    tmpfile = fopen("/etc/config/system.glGammaVal", "r");
    if (tmpfile) {
        double sgi_gamma;

        fgets(tmpline, 80, tmpfile);
        fclose(tmpfile);
        sgi_gamma = atof(tmpline);
        if (sgi_gamma > 0.0)
            LUT_exponent = 1.0 / sgi_gamma;
    }
#elif defined(Macintosh)
    LUT_exponent = 1.8 / 2.61;
#else
    LUT_exponent = 1.0;   /* assume no LUT:  most PCs */
#endif

    /* the defaults above give 1.0, 1.3, 1.5 and 2.2, respectively: */
    default_display_exponent = LUT_exponent * CRT_exponent;


    /* If the user has set the SCREEN_GAMMA environment variable as suggested
     * (somewhat imprecisely) in the libpng documentation, use that; otherwise
     * use the default value we just calculated.  Either way, the user may
     * override this via a command-line option. */

    if ((p = getenv("SCREEN_GAMMA")) != NULL) {
      double exponent = atof(p);

      if (exponent > 0.0)
        default_gamma = 1.0 / exponent;
    }
    
    if (default_gamma == 0.0)
      default_gamma = 1.0 / default_display_exponent;
    
    
    if (type != 5  && type != 6){
      fprintf(stderr, PROGNAME
              ":  only P5, P6 file allowed \n");
      writepng_cleanup(&wpng_info);
      wpng_cleanup();
      return 0;
    }
    wpng_info.pnmtype = type;
    wpng_info.width = width;
    wpng_info.height = height;
    
    if (wpng_info.width <= 0L || wpng_info.height <= 0L ||
        maxval != 255) {
      fprintf(stderr, PROGNAME
              ":  only positive width/height, maxval == 255 allowed \n");
      writepng_cleanup(&wpng_info);
      wpng_cleanup();
      return 0;
    }
      
    wpng_info.sample_depth = 8;  /* <==> maxval 255 */
    wpng_info.outfile = outf;

    /* allocatelibpng stuff, initialize transformations, write pre-IDAT data */
    if ((rc = writepng_init(&wpng_info)) != 0) {
        switch (rc) {
            case 2:
                fprintf(stderr, PROGNAME
                  ":  libpng initialization problem (longjmp)\n");
                break;
            case 4:
                fprintf(stderr, PROGNAME ":  insufficient memory\n");
                break;
            case 11:
                fprintf(stderr, PROGNAME
                  ":  internal logic error (unexpected PNM type)\n");
                break;
            default:
                fprintf(stderr, PROGNAME
                  ":  unknown writepng_init() error\n");
                break;
        }
        writepng_cleanup(&wpng_info);
        wpng_cleanup();
        return 0;
    }

    /* calculate rowbytes on basis of image type; note that this becomes much
     * more complicated if we choose to support PBM type, ASCII PNM types, or
     * 16-bit-per-sample binary data [currently not an official NetPBM type] */

    if (wpng_info.pnmtype == 5)
        rowbytes = wpng_info.width;
    else
        rowbytes = wpng_info.width * 3;

    /* read and write the image, either in its entirety (if writing interlaced
     * PNG) or row by row (if non-interlaced) */

    //fprintf(stderr, "Encoding image data...\n");
    fflush(stderr);

    if (wpng_info.interlaced) {
      long i;

      ulg image_bytes = rowbytes * wpng_info.height;   /* overflow? */
      
      wpng_info.image_data = (uch *)malloc(image_bytes);
      wpng_info.row_pointers = 
        (uch **)malloc(wpng_info.height*sizeof(uch *));
      if (wpng_info.image_data == NULL || wpng_info.row_pointers == NULL) {
        fprintf(stderr, PROGNAME ":insufficient memory for image data\n");
        writepng_cleanup(&wpng_info);
        wpng_cleanup();
        return 0;
      }
      for (i = 0;  i < wpng_info.height;  ++i)
        wpng_info.row_pointers[i] = wpng_info.image_data + i*rowbytes;

      if(data == NULL){
        fprintf(stderr, PROGNAME ":data is null\n");
        return -1;
      }

      memcpy(wpng_info.image_data, data, image_bytes);

      if (writepng_encode_image(&wpng_info) != 0) {
        fprintf(stderr, PROGNAME
                ":libpng problem (longjmp) while writing image data\n");
        writepng_cleanup(&wpng_info);
        wpng_cleanup();
        return 0;
      }
      
    } else /* not interlaced:  write progressively (row by row) */ {
      long j;
      
      wpng_info.image_data = (uch *)malloc(rowbytes);
      if (wpng_info.image_data == NULL) {
        fprintf(stderr, PROGNAME ":  insufficient memory for row data\n");
        writepng_cleanup(&wpng_info);
        wpng_cleanup();
        return 0;
      }
      error = 0;

      for (j = 0L; j < wpng_info.height; j++) {
        
        memcpy(wpng_info.image_data, data + rowbytes*j, rowbytes);

        if (writepng_encode_row(&wpng_info) != 0) {
          fprintf(stderr, PROGNAME
                  ":libpng problem (longjmp) while writing row %ld\n",
                  wpng_info.height-j);
          ++error;
          break;
        }
      }
      if (error) {
        writepng_cleanup(&wpng_info);
        wpng_cleanup();
        return 0;
      }
      if (writepng_encode_finish(&wpng_info) != 0) {
        fprintf(stderr, PROGNAME ":  error on final libpng call\n");
        writepng_cleanup(&wpng_info);
        wpng_cleanup();
        return 0;
      }
    }
    
    /* OK, we're done (successfully):  clean up all resources and quit */
    //fprintf(stderr, "Done.\n");
    fflush(stderr);

    writepng_cleanup(&wpng_info);
    wpng_cleanup();

    return 1;
}

static void wpng_cleanup(void)
{
    if (wpng_info.outfile) {
        wpng_info.outfile = NULL;
    }

    if (wpng_info.image_data) {
        free(wpng_info.image_data);
        wpng_info.image_data = NULL;
    }

    if (wpng_info.row_pointers) {
        free(wpng_info.row_pointers);
        wpng_info.row_pointers = NULL;
    }
}
